/*
    SPDX-FileCopyrightText: 2008 Joris Guisson <joris.guisson@gmail.com>
    SPDX-FileCopyrightText: 2008 Ivan Vasic <ivasic@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/
#include "cachefactory.h"

namespace bt
{
CacheFactory::CacheFactory()
{
}

CacheFactory::~CacheFactory()
{
}

}
