/*
    SPDX-FileCopyrightText: 2005 Joris Guisson <joris.guisson@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/
#include "chunkdownloadinterface.h"

namespace bt
{
ChunkDownloadInterface::ChunkDownloadInterface()
{
}

ChunkDownloadInterface::~ChunkDownloadInterface()
{
}

}
