/*
    SPDX-FileCopyrightText: 2005 Joris Guisson <joris.guisson@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/
#include "piece.h"
#if 0
namespace bt
{

Piece::Piece(Uint32 index, Uint32 off, Uint32 len, PieceDownloader* pd, const Uint8* data)
    : Request(index, off, len, pd), data(data)
{}


Piece::~Piece()
{}


}
#endif
