/*
    SPDX-FileCopyrightText: 2012 Joris Guisson <joris.guisson@gmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "rpcserverinterface.h"

namespace dht
{
RPCServerInterface::RPCServerInterface()
{
}

RPCServerInterface::~RPCServerInterface()
{
}
}
